unit TestMain;

interface

uses
  Windows, SysUtils, Graphics, Controls, Dialogs, Forms, Math, ExtCtrls, Classes;

type

  TForm1 = class;

  TMyShape = class(TObject)
  private
    FOnClick: TNotifyEvent;
    FOnMouseMove: TMouseMoveEvent;
    function GetCanvas: TCanvas;
    function GetClientRect: TRect;
  protected
    FOwner: TForm1;
  public
  	ID: string;
    Left: Integer;
    Top: Integer;
    Width: Integer;
    Height: Integer;
    Color: TColor;
    constructor Create(AOwner: TForm1);
    procedure Draw; virtual; abstract;
    procedure Click;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);
    property Owner: TForm1 read FOwner;
    property Canvas: TCanvas read GetCanvas;
    property ClientRect: TRect read GetClientRect;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
  end;

  TMyShape1 = class(TMyShape)
  public
    procedure Draw; override;
  end;

  TMyShape2 = class(TMyShape)
  public
    procedure Draw; override;
  end;

  TForm1 = class(TForm)
    Timer1: TTimer;
  private
    List: TStringList;
    procedure CREATE_OBJECTS;
    procedure DoClick(Sender: TObject);
    procedure DO_UPDATE(Sender: TObject; var Done: Boolean);
    procedure DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  protected
  	procedure Click; override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure Resize; override;
  public
    constructor Create(AOnwer: TComponent); override;
    destructor Destroy; override;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{ TMyShape }

procedure TMyShape.Click;
begin
	if Assigned(FOnClick) then FOnClick(Self);
end;

constructor TMyShape.Create(AOwner: TForm1);
begin
	inherited Create;
  FOwner := AOwner;
end;

function TMyShape.GetCanvas: TCanvas;
begin
	Result := FOwner.Canvas;
end;

function TMyShape.GetClientRect: TRect;
begin
	Result := Rect(Left, Top, Left+Width-1, Top+Height-1);
end;

procedure TMyShape.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnMouseMove) then FOnMouseMove(Self, Shift, X, Y);
end;

{ TMyShape1 }

procedure TMyShape1.Draw;
begin
	with Canvas do
  begin
    Lock;
    try
      Pen.Color := Color;
      Pen.Width := 2;
      Pen.Style := psSolid;
      Pen.Mode := pmXor;
      Brush.Color := Owner.Color;
      with ClientRect do
        PolyLine([Point(Left,Top),
                  Point(Left,Bottom),
                  Point(Right,Bottom),
                  Point(Right,Top),
                  Point(Left,Top)]);
    finally
      Unlock;
    end;
  end;
end;

{ TMyShape2 }

procedure TMyShape2.Draw;
begin
	with Canvas do
  begin
    Lock;
    try
      Pen.Color := Owner.Color;
      Pen.Width := 2;
      Pen.Style := psSolid;
      Pen.Mode := pmXor;
      Brush.Color := Color;
      Ellipse(ClientRect);
    finally
      Unlock;
    end;
  end;
end;

{ TForm1 }

constructor TForm1.Create(AOnwer: TComponent);
begin
  inherited;
  Color :=clBlack;
  Randomize;
  List := TStringList.Create;
  CREATE_OBJECTS;
  WindowState := wsMaximized;
  Application.OnIdle := DO_UPDATE;
end;

procedure TForm1.CREATE_OBJECTS;
var
	I: Integer;
  O: TMyShape;
begin
	for I := 0 to 99 do
  begin
    case Random(MaxInt) mod 2 of
      0:    O := TMyShape1.Create(Self);
      else  O := TMyShape2.Create(Self);
    end;
    O.ID := Format('%2.2d', [I]);
    O.Color := RGB(Random($FF),Random($FF),Random($FF));
  	O.Width := Max(50,Random(100));
    O.Height := O.Width;
    O.Left := Random(Screen.Width-100);
    O.Top := Random(Screen.Height-100);
    O.OnClick := DoClick;
    O.OnMouseMove := DoMouseMove;
    List.AddObject('', O);
  end;
end;

destructor TForm1.Destroy;
var
	I: Integer;
begin
	if List <> nil then
  begin
    for I := 0 to List.Count-1 do
    	List.Objects[I].Free;
    List.Free;
  end;
  inherited;
end;

procedure TForm1.DoClick(Sender: TObject);
begin
	ShowMessage('TMyShape Object Clicked: ID = ' + TMyShape(Sender).ID);
end;

procedure TForm1.Paint;
var
	I: Integer;
begin
  inherited;
  if List <> nil then
  	for I := 0 to List.Count-1 do
    	TMyShape(List.Objects[I]).Draw;
end;

procedure TForm1.Click;
var
	Pt: TPoint;
	I: Integer;
  O: TMyShape;
begin
	GetCursorPos(Pt);
  Pt := ScreenToClient(Pt);
	for I := 0 to List.Count-1 do
  begin
  	O := TMyShape(List.Objects[I]);
    if PtInRect(O.ClientRect, Pt) then
    begin
    	O.Click;
      Break;
    end;
  end;
  inherited;
end;

procedure TForm1.MouseMove(Shift: TShiftState; X, Y: Integer);
var
	Pt: TPoint;
	I: Integer;
  O: TMyShape;
begin
	GetCursorPos(Pt);
  Pt := ScreenToClient(Pt);
	for I := 0 to List.Count-1 do
  begin
  	O := TMyShape(List.Objects[I]);
    if PtInRect(O.ClientRect, Pt) then
    begin
    	O.MouseMove(Shift, X, Y);
      Break;
    end;
  end;
  inherited;
end;

procedure TForm1.Resize;
begin
	inherited;
  Invalidate;
end;

procedure TForm1.DO_UPDATE(Sender: TObject; var Done: Boolean);
var
  I, J, W, H: Integer;
  O: TMyShape;
begin
  if List <> nil then
  begin
    W := Screen.Width-100;
    H := Screen.Height-100;
    for I := 0 to List.Count div 4 do
    begin
      J := Random(List.Count);
      O := TMyShape(List.Objects[J]);
      O.Draw;
      O.Left := (O.Left + Random(7)-3 + W) mod W;
      O.Top := (O.Top + Random(7)-3 + H) mod H;
      O.Draw;
    end;
  end;
  Done := False;
  Application.ProcessMessages;
  Sleep(100);
end;


procedure TForm1.DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  Caption := 'MouseMove: ID = ' + TMyShape(Sender).ID;
  Caption := Caption + ', X=' + IntToStr(X) + ', Y=' + IntToStr(Y);
end;

end.
